/******************************************************************************
**
** FILE NAME    : amazon_se_pmu.c
** PROJECT      : Amazon_se
** MODULES      : PMU
**
** DATE         : 1 Jan 2006
** AUTHOR       : Huang Xiaogang
** DESCRIPTION  : Amazon_se Power Management Unit driver
** COPYRIGHT    :       Copyright (c) 2006
**                      Infineon Technologies AG
**                      Am Campeon 1-12, 85579 Neubiberg, Germany
**
**    This program is free software; you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation; either version 2 of the License, or
**    (at your option) any later version.
**
** HISTORY
** $Date        $Author         $Comment
** 01 Jan 2006  Huang Xiaogang  modification & verification on Amazon_se chip
*******************************************************************************/
#include<linux/module.h>
#include<linux/init.h>
#include<linux/proc_fs.h>
#include<linux/kdev_t.h>
#include<asm/atomic.h>
#include<linux/fs.h>

#include<asm/amazon_se/amazon_se.h>
#include<asm/amazon_se/amazon_se_pmu.h>
#include<asm/uaccess.h>

atomic_t amazon_se_pmu_in_use;
static pmu_dev *amazon_se_pmu_dev;

int pmu_set(int power_unit, int value)
{
        u32 cr = 0;
        
        cr = *AMAZON_SE_PMU_PWDCR;
        
        if(value)
                cr |= 1 << power_unit;
        else
                cr &= ((~(1 << power_unit)) & 0x03FFFFFF);/* mask 0 - 25 */
        
        *AMAZON_SE_PMU_PWDCR = cr;
//      printk("AMAZON_SE_PMU_PWDCR = 0x%08x\n", cr);
        return 1;
}

void pmu_set_pmu_pwdcr(int istatus)
{
        *AMAZON_SE_PMU_PWDCR = istatus;
}

int pmu_open(struct inode *inode, struct file *filp)
{
        if (atomic_dec_and_test(&amazon_se_pmu_in_use)) {
                atomic_inc(&amazon_se_pmu_in_use);
                return -EBUSY;
        }
        return 0;
}

int pmu_release(struct inode *inode, struct file *filp)
{
        atomic_inc(&amazon_se_pmu_in_use);
        return 0;
}

static int pmu_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg)
{
        int value_c, value_s, en;

        printk("cmd=0x%x\n", cmd);

        switch(cmd)
        {
        case AMAZON_SE_PMU_IOC_SET_PWDCR: 
                __get_user(value_c, (int *)arg);
                pmu_set_pmu_pwdcr(value_c);
                break;
        case AMAZON_SE_PMU_IOC_GET_STATUS:
                __put_user((int)(*AMAZON_SE_PMU_SR), (int*)arg);
                break;
        case AMAZON_SE_PMU_IOC_SET_USBPHY:
        case AMAZON_SE_PMU_IOC_SET_FPI2:    
        case AMAZON_SE_PMU_IOC_SET_SDIO: 
        case AMAZON_SE_PMU_IOC_SET_DMA:
        case AMAZON_SE_PMU_IOC_SET_USB:        
        case AMAZON_SE_PMU_IOC_SET_EPHY:    
        case AMAZON_SE_PMU_IOC_SET_SPI:    
        case AMAZON_SE_PMU_IOC_SET_DSL:  
        case AMAZON_SE_PMU_IOC_SET_EBU:         
        case AMAZON_SE_PMU_IOC_SET_LEDC:      
        case AMAZON_SE_PMU_IOC_SET_GPTC:    
        case AMAZON_SE_PMU_IOC_SET_PPE:   
        case AMAZON_SE_PMU_IOC_SET_FPI0:     
        case AMAZON_SE_PMU_IOC_SET_AHB:    
        case AMAZON_SE_PMU_IOC_SET_UART1:  
        case AMAZON_SE_PMU_IOC_SET_WDT0: 
        case AMAZON_SE_PMU_IOC_SET_PPE_TC:
        case AMAZON_SE_PMU_IOC_SET_PPE_ENET0:           
                __get_user(en, (int*)arg);   
                pmu_set((cmd & 0xff), en);
                break;
        default:
                return -ENOTTY;
        }
        
        return 0;
}

static struct file_operations pmu_fops =
{
        owner:    THIS_MODULE,
        open:     pmu_open,
        release:  pmu_release,
        ioctl:    pmu_ioctl,
};

static int amazon_se_pmu_proc_read(char *buf, char **start, off_t offset, int count, int *eof, void *data)
{
        int len=0;
        len+=sprintf(buf+len,"AMAZON_SE_PMU_PROC_READ\n");

        len+=sprintf(buf+len,"AMAZON_SE_PMU_PWDCR(0x%08x) : 0x%08x\n", AMAZON_SE_PMU_PWDCR, *AMAZON_SE_PMU_PWDCR);
        len+=sprintf(buf+len,"AMAZON_SE_PMU_SR(0x%08x)    : 0x%08x\n", AMAZON_SE_PMU_SR, *AMAZON_SE_PMU_SR);

        *eof = 1;
        return len;
}

int __init amazon_se_pmu_init_module(void)
{
        int result;
        dev_t dev = 0;

        printk(KERN_INFO"Amazon_se PMU driver v0.3\n");

        amazon_se_pmu_dev = (pmu_dev*)kmalloc(sizeof(pmu_dev), GFP_KERNEL);
        if(amazon_se_pmu_dev == NULL)
                return -ENOMEM;
        memset(amazon_se_pmu_dev, 0, sizeof(pmu_dev));

        strcpy(amazon_se_pmu_dev->name, "amazon_se_pmu");
        amazon_se_pmu_dev->minor = 0;
        result = register_chrdev(0, amazon_se_pmu_dev->name, &pmu_fops);
        amazon_se_pmu_dev->major = result;
        
        create_proc_read_entry("amazon_se_pmu", 0, NULL, amazon_se_pmu_proc_read, NULL);
        atomic_set(&amazon_se_pmu_in_use,0);

        return 0;
}

void amazon_se_pmu_cleanup_module(void)
{
        unregister_chrdev(amazon_se_pmu_dev->major, amazon_se_pmu_dev->name);
        remove_proc_entry("amazon_se_pmu", NULL);  
}

MODULE_AUTHOR("Huang Xiaogang <xiaogang.huang@infineon.com>");
MODULE_DESCRIPTION("Amazon_se Power Management Unit driver");
MODULE_LICENSE("GPL");

module_init(amazon_se_pmu_init_module);
module_exit(amazon_se_pmu_cleanup_module);

